From: "Zhang Yi" <wetpzy@gmail.com>
To: <linux-kernel@vger.kernel.org>
Cc: "'Peter Zijlstra'" <peterz@infradead.org>,
"'Darren Hart'" <dvhart@linux.intel.com>,
"'Thomas Gleixner'" <tglx@linutronix.de>,
"'Ingo Molnar'" <mingo@kernel.org>,
"'Dave Hansen'" <dave.hansen@linux.intel.com>,
<zhang.yi20@zte.com.cn>
Subject: [PATCH] futex: bugfix for futex-key conflict when futex use hugepage
Date: Wed, 24 Apr 2013 22:27:32 +0800 [thread overview]
Message-ID: <000a01ce40f7$e9f4d700$bdde8500$@com> (raw)
Hi all,
I reworked the patch base on your advices.
For the line-wrapped bug before, I use this mailbox to send the mail .
Signed-off-by: Zhang Yi <zhang.yi20@zte.com.cn>
Tested-by: Ma Chenggong <ma.chenggong@zte.com.cn>
Reviewed-by: Liu Dong <liu.dong3@zte.com.cn>
Reviewed-by: Cui Yunfeng <cui.yunfeng@zte.com.cn>
Reviewed-by: Lu Zhongjun <lu.zhongjun@zte.com.cn>
Reviewed-by: Jiang Biao <jiang.biao2@zte.com.cn>
diff -uprN orig/linux3.9-rc7/include/linux/futex.h new/linux3.9-rc7/include/linux/futex.h
--- orig/linux3.9-rc7/include/linux/futex.h 2013-04-15 00:45:16.000000000 +0000
+++ new/linux3.9-rc7/include/linux/futex.h 2013-04-19 16:33:58.725880000 +0000
@@ -19,7 +19,7 @@ handle_futex_death(u32 __user *uaddr, st
* The key type depends on whether it's a shared or private mapping.
* Don't rearrange members without looking at hash_futex().
*
- * offset is aligned to a multiple of sizeof(u32) (== 4) by definition.
+ * There are three cmponents in offset:
* We use the two low order bits of offset to tell what is the kind of key :
* 00 : Private process futex (PTHREAD_PROCESS_PRIVATE)
* (no reference on an inode or mm)
@@ -27,6 +27,9 @@ handle_futex_death(u32 __user *uaddr, st
* mapped on a file (reference on the underlying inode)
* 10 : Shared futex (PTHREAD_PROCESS_SHARED)
* (but private mapping on an mm, and reference taken on it)
+ * Bits 2 to (PAGE_SHIFT-1) indicates the offset of futex in its page.
+ * The rest hign order bits indicates the index if the page is a
+ * subpage of a compound page.
*/
#define FUT_OFF_INODE 1 /* We set bit 0 if key has a reference on inode */
@@ -36,17 +39,17 @@ union futex_key {
struct {
unsigned long pgoff;
struct inode *inode;
- int offset;
+ long offset;
} shared;
struct {
unsigned long address;
struct mm_struct *mm;
- int offset;
+ long offset;
} private;
struct {
unsigned long word;
void *ptr;
- int offset;
+ long offset;
} both;
};
diff -uprN orig/linux3.9-rc7/kernel/futex.c new/linux3.9-rc7/kernel/futex.c
--- orig/linux3.9-rc7/kernel/futex.c 2013-04-15 00:45:16.000000000 +0000
+++ new/linux3.9-rc7/kernel/futex.c 2013-04-19 16:24:05.629143000 +0000
@@ -215,6 +215,22 @@ static void drop_futex_key_refs(union fu
}
}
+/*
+* Get subpage index in compound page, for futex_key.
+*/
+static inline int get_page_compound_index(struct page *page)
+{
+ struct page *head_page;
+ if (PageHead(page))
+ return 0;
+
+ head_page = compound_head(page);
+ if (compound_order(head_page) >= MAX_ORDER)
+ return page_to_pfn(page) - page_to_pfn(head_page);
+ else
+ return page - compound_head(page);
+}
+
/**
* get_futex_key() - Get parameters which are the keys for a futex
* @uaddr: virtual address of the futex
@@ -228,7 +244,8 @@ static void drop_futex_key_refs(union fu
* The key words are stored in *key on success.
*
* For shared mappings, it's (page->index, file_inode(vma->vm_file),
* offset_within_page). For private mappings, it's (uaddr, current->mm).
* We can usually work out the index without swapping in the page.
*
* lock_page() might sleep, the caller should not hold a spinlock.
@@ -239,7 +256,7 @@ get_futex_key(u32 __user *uaddr, int fsh
unsigned long address = (unsigned long)uaddr;
struct mm_struct *mm = current->mm;
struct page *page, *page_head;
- int err, ro = 0;
+ int err, ro = 0, comp_idx = 0;
/*
* The futex address must be "naturally" aligned.
@@ -299,6 +316,7 @@ again:
* freed from under us.
*/
if (page != page_head) {
+ comp_idx = get_page_compound_index(page);
get_page(page_head);
put_page(page);
}
@@ -311,6 +329,7 @@ again:
#else
page_head = compound_head(page);
if (page != page_head) {
+ comp_idx = get_page_compound_index(page);
get_page(page_head);
put_page(page);
}
@@ -363,7 +382,9 @@ again:
key->private.mm = mm;
key->private.address = address;
} else {
- key->both.offset |= FUT_OFF_INODE; /* inode-based key */
+ /* inode-based key */
+ key->both.offset |= ((long)comp_idx << PAGE_SHIFT)
+ | FUT_OFF_INODE;
key->shared.inode = page_head->mapping->host;
key->shared.pgoff = page_head->index;
}
next reply other threads:[~2013-04-24 15:32 UTC|newest]
Thread overview: 35+ messages / expand[flat|nested] mbox.gz Atom feed top
2013-04-24 14:27 Zhang Yi [this message]
-- strict thread matches above, loose matches on Subject: below --
2013-05-15 13:57 [PATCH] futex: bugfix for futex-key conflict when futex use hugepage Zhang Yi
2013-05-15 14:20 ` Mel Gorman
2013-05-16 1:16 ` zhang.yi20
2013-05-16 1:30 ` Darren Hart
2013-05-07 12:43 Zhang Yi
2013-04-26 12:13 Zhang Yi
2013-04-26 18:26 ` Thomas Gleixner
2013-05-07 12:23 ` Zhang Yi
2013-05-07 15:20 ` Mel Gorman
2013-05-07 15:24 ` Thomas Gleixner
2013-05-07 15:54 ` Mel Gorman
2013-05-07 12:34 ` Zhang Yi
2013-04-24 13:58 Zhang Yi
2013-04-25 20:52 ` Thomas Gleixner
2013-04-17 9:55 zhang.yi20
2013-04-17 14:18 ` Darren Hart
2013-04-17 14:18 ` Darren Hart
2013-04-17 15:26 ` Dave Hansen
2013-04-17 15:26 ` Dave Hansen
2013-04-17 15:51 ` Darren Hart
2013-04-17 15:51 ` Darren Hart
2013-04-18 8:05 ` zhang.yi20
2013-04-18 14:34 ` Darren Hart
2013-04-18 14:34 ` Darren Hart
2013-04-19 2:13 ` zhang.yi20
2013-04-19 2:42 ` Darren Hart
2013-04-19 2:42 ` Darren Hart
2013-04-19 2:45 ` Darren Hart
2013-04-19 2:45 ` Darren Hart
2013-04-16 3:37 zhang.yi20
2013-04-16 3:37 ` zhang.yi20
2013-04-16 17:57 ` Darren Hart
2013-04-16 17:57 ` Darren Hart
2013-04-16 18:37 ` Dave Hansen
2013-04-16 18:37 ` Dave Hansen
2013-04-16 18:47 ` Darren Hart
2013-04-16 18:47 ` Darren Hart
2013-04-08 6:34 jiang.biao2
Reply instructions:
You may reply publicly to this message via plain-text email
using any one of the following methods:
* Save the following mbox file, import it into your mail client,
and reply-to-all from there: mbox
Avoid top-posting and favor interleaved quoting:
https://en.wikipedia.org/wiki/Posting_style#Interleaved_style
* Reply using the --to, --cc, and --in-reply-to
switches of git-send-email(1):
git send-email \
--in-reply-to='000a01ce40f7$e9f4d700$bdde8500$@com' \
--to=wetpzy@gmail.com \
--cc=dave.hansen@linux.intel.com \
--cc=dvhart@linux.intel.com \
--cc=linux-kernel@vger.kernel.org \
--cc=mingo@kernel.org \
--cc=peterz@infradead.org \
--cc=tglx@linutronix.de \
--cc=zhang.yi20@zte.com.cn \
/path/to/YOUR_REPLY
https://kernel.org/pub/software/scm/git/docs/git-send-email.html
* If your mail client supports setting the In-Reply-To header
@@ -228,7 +244,8 @@ static void drop_futex_key_refs(union fu
*
* For shared mappings, it's (page->index, file_inode(vma->vm_file),
- * offset_within_page). For private mappings, it's (uaddr, current->mm).
+ * page_compound_index and offset_within_page).
+ * For private mappings, it's (uaddr, current->mm).
* We can usually work out the index without swapping in the page.
*
via mailto: links, try the mailto: link
Be sure your reply has a Subject: header at the top and a blank line
before the message body.
This is an external index of several public inboxes,
see mirroring instructions on how to clone and mirror
all data and code used by this external index.